/* 
 * FILE --- rbf.cc
 *
 * PURPOSE --- Spurce file for a class that implements a Radial Basis 
 * Function (RBF) neural network (NN) with gaussian nodes. 
 * The user specifies the number of inputs, the number of outputs and the 
 * number of the neurons (i.e., the gaussian units). The user needs to provide
 * the centers and the spreads og the gaussian nodes. Moreover, s/he can either 
 * specify the output parameters or let them be choosen randomly within the user defined range.
 * The utility has a build-in adaptation capability of the output parameters
 * based on the Normalized Gradient Method or the so called Projection Algorithm.
 * Therefore it can be used in both direct and indirect adaptive control schemes as
 * a function approximator.
 *
 * This utility uses the matrix tool for some matrix maltiplications.
 */


// Include Files
#include <math.h>
#include "rbf.hh"
#include "random.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif


/*************************************************************/

RBF::RBF(int _num_of_inputs, int _num_of_outputs, int _num_of_neurons)
{
  // No error at start
  rbf_err = RBF_NO_ERROR;
  num_of_inputs = _num_of_inputs;
  num_of_outputs = _num_of_outputs;
  num_of_neurons = _num_of_neurons;

  // Allocate memory 
  allocate_memory();

  // If any error then report
  if (rbf_err != RBF_NO_ERROR)
    {
#ifdef NO_RCS
      printf("RBF: ERROR --- Not Enought Memory!!!\n");
      printf("The RBF network will not function correctly ...\n");
#else
      rcs_print("RBF: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The RBF network will not function correctly ...\n");
#endif
    }

  // Initialize the output parameters with random numbers b/n [-1.0, 1.0]
  Randomize();
  set_output_params(-1.0, 1.0);

  // Initialize all the spreads to 1.0
  set_spreads(1.0);

  // Initialize all the centers to 0.0
  set_centers(0.0);

  // Initialize the adaptation gain to 1.0
  eta = 1.0;

  // Phi norm is zero initially
  phi_norm = 0.0;
}
  

RBF::~RBF()
{
  // Free the allocated memory
  free_memory();
}


void RBF::allocate_memory(void)
{
  // Create the vector of centers of the neurons
  if ((centers = new double*[num_of_neurons]) == NULL)
    rbf_err = RBF_MEM_ERROR;
  
  // Creater each center
  for (int ii=0; ii < num_of_neurons; ii++)
    {
      if ((centers[ii] = new double[num_of_inputs]) == NULL)
	rbf_err = RBF_MEM_ERROR;
    }

  // Create the vector of spreads of the neurons
  if ((spreads = new double*[num_of_neurons]) == NULL)
    rbf_err = RBF_MEM_ERROR;
  
  // Creater each spread
  for (int ii=0; ii < num_of_neurons; ii++)
    {
      if ((spreads[ii] = new double[num_of_inputs]) == NULL)
	rbf_err = RBF_MEM_ERROR;
    }

  // Create the vector of output parametrs
  theta = new matrix(num_of_outputs, num_of_neurons);
  if (theta->matrix_err != MATRIX_NO_ERROR)
    {
      rbf_err = RBF_MEM_ERROR;
    }

  // Create the vector of function values of the neurons
  phi = new matrix(num_of_neurons, 1);
  if (phi->matrix_err != MATRIX_NO_ERROR)
    {
      rbf_err = RBF_MEM_ERROR;
    }

  // Create output vector y
  y = new matrix(num_of_outputs, 1);
  if (y->matrix_err != MATRIX_NO_ERROR)
    {
      rbf_err = RBF_MEM_ERROR;
    }

  // Create error vector e
  e = new matrix(num_of_outputs, 1);
  if (e->matrix_err != MATRIX_NO_ERROR)
    {
      rbf_err = RBF_MEM_ERROR;
    }
}


void RBF::free_memory(void)
{
  for (int ii=0; ii < num_of_neurons; ii++)
    {
      delete [] centers[ii];
      delete [] spreads[ii];
    }
  delete [] centers;
  delete [] spreads;

  delete theta;
  delete phi;
 
  delete y;
  delete e;
}

void RBF::set_spreads(double _spreads)
{
  int ii, jj;  // Index variales for the loops

  // Assign all the spreads for a given input to the same value
  // Different inputs can have can have different spreads
  for (ii=0; ii < num_of_neurons; ii++)
    {
      for (jj=0; jj < num_of_inputs; jj++)
	{
	  // Make sure that the spreads are positive
	  spreads[ii][jj] = fabs(_spreads);
	}
    }
}


void RBF::set_spreads(double *_spreads)
{
  int ii, jj;  // Index variales for the loops

  // Assign all the spreads for a given input to the same value
  // Different inputs can have can have different spreads
  for (ii=0; ii < num_of_neurons; ii++)
    {
      for (jj=0; jj < num_of_inputs; jj++)
	{
	  // Make sure that the spreads are positive
	  spreads[ii][jj] = fabs(_spreads[jj]);
	}
    }
}


void RBF::set_spreads(double **_spreads)
{
  int ii, jj;  // Index variales for the loops

  // Assign each spread to a (possibly) different value
  for (ii=0; ii < num_of_neurons; ii++)
    {
      for (jj=0; jj < num_of_inputs; jj++)
	{
	  // Make sure that the spreads are positive
	  spreads[ii][jj] = fabs(_spreads[ii][jj]);
	}
    }
}


void RBF::set_centers(double _centers)
{
  int ii, jj;  // Index variales for the loops

  // Assign each center
  for (ii = 0; ii < num_of_neurons; ii++)
    {
      for (jj = 0; jj < num_of_inputs; jj++)
	{
	  centers[ii][jj] = _centers;
	}
    }
}


void RBF::set_centers(double **_centers)
{
  int ii, jj;  // Index variales for the loops

  // Assign each center
  for (ii = 0; ii < num_of_neurons; ii++)
    {
      for (jj = 0; jj < num_of_inputs; jj++)
	{
	  centers[ii][jj] = _centers[ii][jj];
	}
    }
}


void RBF::set_output_params(double _low, double _high)
{
  int ii, jj;  // Index variales for the loops
  double rand_val;  // Temporary variable

  // Set the output parametrs random values b/n [_low, _high]
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      for (jj = 0; jj < num_of_neurons; jj++)
        {
	   // Generate a random number b/n [_low, _high]
	   rand_val = Rand_Double(_low, _high);
	   
	   // Assign the element to the generated value
	   theta->set_element(ii, jj, rand_val);
	 }
    }
}

void RBF::set_output_params(const double _out_params)
{
  // Set the output gain matrix to the specified value
  theta->set_matrix(_out_params);
}


void RBF::set_output_params(const double *_out_params)
{
  // Set the output gain matrix to the specified value
  theta->set_matrix(_out_params);
}


void RBF::set_output_params(const double **_out_params)
{
  // Set the output gain matrix to the specified value
  theta->set_matrix(_out_params);
}


void RBF::control(const double *input, double *_output)
{
  // Calculate the output of the activation functions
  get_activation(input);

  // Calculate the output of the RBF
  (*y) = (*theta) * (*phi);

  // Return the output
  get_output(_output);
}

void RBF::update_output_params(const double *error)
{
  // Use matrix operations
  e->set_col(0, error);

  // Update the parameters based on the Normilized Gradient Method
  // or the so called Projection Algorithm
  (*theta) = (*theta) + (eta/(1.0 + phi_norm)) * (*e) * transpose((*phi));
}


void RBF::update_output_params(const double *error, double norm2)
{
  // Use matrix operations
  e->set_col(0, error);

  // Update the parameters based on the Normilized Gradient Method
  // or the so called Projection Algorithm
  (*theta) = (*theta) + (eta/(1.0 + norm2 + phi_norm)) * (*e) * transpose((*phi));
}


void RBF::update_output_params(const double *error, double norm2, double uu)
{
  // Use matrix operations
  e->set_col(0, error);

  // Update the parameters based on the Normilized Gradient Method
  // or the so called Projection Algorithm
  (*theta) = (*theta) 
    + ((eta*uu)/(1.0 + norm2 + uu*uu*phi_norm)) * (*e) * transpose((*phi));
}


void RBF::bound_output_params(double _low, double _high)
{
  int ii, jj;  // Index variales for the loops

  // Saturates the output parametrs values b/n [_low, _high]
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      for (jj = 0; jj < num_of_neurons; jj++)
        {
	  if (theta->get_element(ii, jj) < _low)
	    {
	      // Assign the element to the generated value
	      theta->set_element(ii, jj, _low);
	    }
	  else if (theta->get_element(ii, jj) > _high)
	    {
	      // Assign the element to the generated value
	      theta->set_element(ii, jj, _high);
	    }
	 }
    }
}


void RBF::get_activation(const double *input)
{
  int ii, jj;  // Index variales for the loops
  double func_val;  // Temporary variable

  // Calculate the activation of each neuron
  for(ii = 0; ii < num_of_neurons; ii++)
    {
      func_val = 1.0;
      for (jj=0; jj < num_of_inputs; jj++)
	{
	  // Add the contribution of the next input to the output of the neuron
	  func_val *= exp( -pow(((input[jj] - centers[ii][jj])/spreads[ii][jj]), 2.0) );
	}

      // Assign the output to the vector
      phi->set_element(ii, 0, func_val);
    }

  // Calculate the norm of phi (needed in indirect adaptive control)
  phi_norm = (transpose((*phi)) * (*phi)).get_element(0, 0);
}

void RBF::get_output(double *_output)
{
  // Return the output
  for(int ii = 0; ii < num_of_outputs; ii++)
    {
      _output[ii] = y->get_element(ii, 0);
    }
}



