/* 
 * FILE --- rbf_ex1.cc
 *
 * PURPOSE --- To provide an example use of the class for 4'th
 * order Runge-Kutta simulation, RK4_SIM, and the class for implementing
 * n-input m-output Radial Basis Function (RBF) neural network (NN).
 *
 * Implements the main function of a class for performing a simulation of
 * a simple cargo ship whose steering is controlled by an RBF controller.
 *
 */

// Include Files
#include <stdlib.h>
#include <stdio.h>

#include "rbf.hh"
#include "tanker_sim.hh"

// Local Definitions and Macros
#define PI 3.14159265359
#define NUM_GRID 11
#define NUM_NEURONS NUM_GRID*NUM_GRID

void main(void) 
{
  double Ts = 10.0;             // Sampling time
  double t0 = 0.0;              // Initial time 
  double tmax = 4000.0;         // Total Simulation period

  double y[3] = {0.0, 0.0, 0.0};  // Initial conditions    

  // Parameters of the RBF
  double theta[NUM_NEURONS]; // Output parameters
  double **centers;  // Centers of the neurons
  double spreads[2] = {0.7*(PI/11.0), 0.7*(0.02/11.0)};    // Spreads of the neurons

  // # of simulation steps to be performed by the RK4 algorith
  // in one sampling interval
  int sim_per_cycle = 10;  

  // Create the tanker object
  TANKER_SIM my_tanker(Ts,sim_per_cycle,t0,(double *)y);

  // Create the RBF object
  RBF my_rbf_ctrl(2, 1, NUM_NEURONS);

  double ctrl = 0.0;             // Control variable
  double error[2] = {0.0, 0.0};  // Error variable
  double error_old = 0.0;  // Error variable
  double ref = 0.0;  // Reference input variable
  double temp1, temp2;

  FILE *fp;  // To save data

  if (!(fp = fopen("tanker_rbf_out.dat","w"))) {
    printf("Can't open file.\n");
    exit(1);
  }

  // Allocate memory for the spreads
  centers = new double*[NUM_NEURONS];
  for(int ii = 0; ii < NUM_NEURONS; ii++)
    {
      centers[ii] = new double[2];
    }

  // Initialize the the spreads
  temp1 = -PI/2.0;
  temp2 = -0.01;
  for(int ii = 0; ii < NUM_NEURONS; ii++)
    {
      centers[ii][0] = temp1;
      centers[ii][1] = temp2;

      temp1 += (PI/10.0);
      if ((ii % NUM_GRID) == 10)
	{
	  temp1 = -PI/2.0;
	  temp2 += (0.02/10.0);
	}
    }

  // Initialize the output parameters
   for(int ii = 0; ii < NUM_GRID; ii++)
    {
      for(int jj = 0; jj < NUM_GRID; jj++)
	{
	  // Assign each parameter
	  theta[ii+jj*NUM_GRID] = -((1.0/10.0)*(200.0*(PI/180.0))*(ii - 5.0)
				    +(1.0/10.0)*(200.0*(PI/180.0))*(jj - 5.0));

	  // Saturate between max and min possible inputs to the plant
	  if (theta[ii+jj*NUM_GRID] < -80.0*(PI/180.0))
	    {
	      theta[ii+jj*NUM_GRID] = -80.0*(PI/180.0); 
	    }
	  else if (theta[ii+jj*NUM_GRID] > 80.0*(PI/180.0))
	    {
	      theta[ii+jj*NUM_GRID] = 80.0*(PI/180.0);
	    }
	}
    }

  // Set the RBF NN controller parameters
  my_rbf_ctrl.set_spreads(spreads);
  my_rbf_ctrl.set_centers(centers);
  my_rbf_ctrl.set_output_params(theta);

  /* Simulate for tmax seconds. */
  while(my_tanker.current_time < tmax) 
    {
      // Update the reference input
      if (my_tanker.current_time < 100.0)
  	ref = 0;   // Request heading of 0 deg
      else if (my_tanker.current_time < 2000.0)
 	ref = 45.0*(PI/180.0); // Request heading of 45 deg
      else
  	ref = 0;   // Request heading of 0 deg
      
      // Calculate the input to the controller
      error[0] = (ref -  my_tanker.plant_state[0]);
      error[1] = (error[0] - error_old)/Ts;

      // Calculate the control input
      my_rbf_ctrl.control(error, &ctrl);
      
      // Simulate the tanker ship dynamics
      my_tanker.simulate(ctrl);
      
      // Save the data
      fprintf(fp,"%f %f %f %f %f %f \n", my_tanker.current_time, ref,
      	      my_tanker.plant_state[0], my_tanker.plant_state[1],
      	      my_tanker.plant_state[2], my_tanker.ctrl_input[0]);

      // Save the error value
      error_old = error[0];
    }
  
  
  // The code here can be used to view the surface of the function that the RBF implements
  //  error[0] = -80.0*PI/180.0;
  //  for(int ii = 0; ii < 160; ii++)
  //    {
  //      error[1] = -80.0*0.02/180.0;
  //      for (int jj = 0; jj < 160; jj++)
  //  	{
  //          my_rbf_ctrl.control(error, &ctrl);
  //	  fprintf(fp,"%f %f %f \n", error[0], error[1], ctrl);
  //	  error[1] += (2.0*80.0*0.02)/(180.0*160.0);
  //    	}
  //      error[0] += (2.0*80.0*PI)/(180.0*160.0);
  //    }
  
  fclose(fp);
  
  
    for(int ii = 0; ii < NUM_NEURONS; ii++)
    {
      delete [] centers[ii];
    }
    delete [] centers;

  printf("Simulation completed ...\n");
}


