/* 
 * FILE --- rbf_ex2.cc
 *
 * PURPOSE --- To provide an example use of the class for 4'th
 * order Runge-Kutta simulation, RK4_SIM, and the class for implementing
 * n-input m-output adaptive Radial Basis Function (RBF) neural network (NN).
 *
 * Implements the main function of a class for performing a simulation of
 * a simple surge tank level regulation controlled by a direct adaptive RBF controller.
 *
 */

// Include Files
#include <stdlib.h>
#include <stdio.h>

#include "rbf.hh"
#include "tank_sim.hh"

// Local Definitions and Macros
#define PI 3.14159265359
#define NUM_GRID 10
#define NUM_NEURONS NUM_GRID*NUM_GRID

void main(void) 
{
  double Ts = 0.1;             // Sampling time
  double t0 = 0.0;              // Initial time 
  double tmax = 1000.0;         // Total Simulation period

  double y[1] = {1.0};  // Initial conditions    

  // Parameters of the RBF
  double **centers;  // Centers of the neurons
  double spreads = 1.0;  // Spreads of the neurons

  // # of simulation steps to be performed by the RK4 algorith
  // in one sampling interval
  int sim_per_cycle = 2;  

  // Create the tank object
  TANK_SIM my_tank(Ts,sim_per_cycle,t0,(double *)y);

  // Create the RBF object
  RBF my_rbf_ctrl(2, 1, NUM_NEURONS);

  double ctrl = 0.0;             // Control variable
  double input_to_rbf[2] = {0.0, 0.0};  // Error variable
  double error = 0.0;  // Error variable
  double epsilon = 0.001;  // Deadband variable
  double ref = 0.0;  // Reference input variable
  double eta = 1.25; // Adaptation gain
  double temp1, temp2;

  FILE *fp;  // To save data

  if (!(fp = fopen("tank_rbf_dir_out.dat","w"))) {
    printf("Can't open file.\n");
    exit(1);
  }

  // Allocate memory for the spreads
  centers = new double*[NUM_NEURONS];
  for(int ii = 0; ii < NUM_NEURONS; ii++)
    {
      centers[ii] = new double[2];
    }

  // Initialize the the spreads
  temp1 = 0.0;
  temp2 = 0.0;
  for(int ii = 0; ii < NUM_NEURONS; ii++)
    {
      centers[ii][0] = temp1;
      centers[ii][1] = temp2;

      temp1 += 1.0;
      if ((ii % NUM_GRID) == 9)
	{
	  temp1 = 0.0;
	  temp2 += 1.0;
	}
    }

  // Set the RBF NN controller parameters
  my_rbf_ctrl.set_spreads(spreads); // all the spreads to 1.0
  my_rbf_ctrl.set_centers(centers); 
  my_rbf_ctrl.set_output_params(0.0); // all output parameters to 0.0

  // Set the adaptation gain 
  my_rbf_ctrl.set_eta(eta);

  /* Simulate for tmax seconds. */
  while(my_tank.current_time < tmax) 
    {
      // Update the reference input
      if (my_tank.current_time < 100.0)
  	ref = 3.25;   // Request level 3.25
      else if (my_tank.current_time < 300.0)
 	ref = 0.25; // Request level of 0.25 
      else if (my_tank.current_time < 500.0)
 	ref = 3.25; // Request level of 3.25 
      else if (my_tank.current_time < 700.0)
 	ref = 0.25; // Request level of 0.25 
      else if (my_tank.current_time < 900.0)
 	ref = 3.25; // Request level of 3.25 
      else
  	ref = 3.0;   // Request level of 3.0
      
      // Calculate the input to the controller
      input_to_rbf[0] = my_tank.plant_state[0];
      input_to_rbf[1] = ref;

      // Calculate the control input
      my_rbf_ctrl.control(input_to_rbf, &ctrl);
      
      // Simulate the tank ship dynamics
      my_tank.simulate(ctrl);

      // Calculate the error
      error = ref - my_tank.plant_state[0];

      // Pass the error through a deadband
      if (error > epsilon)
	{
	  error -= epsilon;
	}
      else if (error < - epsilon)
	{
	  error += epsilon;
	}

      // Update the output parameters of the controller
      my_rbf_ctrl.update_output_params(&error);

      // Save the data
      fprintf(fp,"%f %f %f %f \n", my_tank.current_time, ref,
      	      my_tank.plant_state[0], my_tank.ctrl_input[0]);
    }
  
  
  // The code here can be used to view the surface of the function that the RBF implements
  //  error[0] = 0.0;
  //  for(int ii = 0; ii < 160; ii++)
  //    {
  //      error[1] = 0.0;
  //      for (int jj = 0; jj < 160; jj++)
  //  	{
  //          my_rbf_ctrl.control(error, &ctrl);
  //	  fprintf(fp,"%f %f %f \n", error[0], error[1], ctrl);
  //	  error[1] += 10.0/160.0;
  //    	}
  //      error[0] += 10.0/160.0;
  //    }
  
  fclose(fp);
  
  
  for(int ii = 0; ii < NUM_NEURONS; ii++)
    {
      delete [] centers[ii];
    }
  delete [] centers;
  
  printf("Simulation completed ...\n");
}


